import asyncio
import sys
import time

import aioblescan
from crownstone_core.util.LogUtil import tfs

counter = 0
prev = time.time()
start = time.time()



class AioScanner:

    def __init__(self, hciIndex = 0):
        self.event_loop = None
        self.bluetoothControl = None
        self.connection = None
        self.timeRequestStart = 0
        self.eventReceived = False
        self.hciIndex = hciIndex
        self.delegate = None

        self.scanRunning = False

        self.scanDuration = 0

    def withDelegate(self, delegate):
        self.delegate = delegate
        return self

    def start(self, duration):
        self.scanRunning = True
        self.scanDuration = duration
        self.scan()

    def stop(self):
        self.scanRunning = False

    def scan(self, attempt = 0):
        print(tfs(), "Attempt Scanning")
        self.eventReceived = False
        event_loop = asyncio.new_event_loop()
        bluetoothSocket = aioblescan.create_bt_socket(self.hciIndex)
        transportProcess = event_loop._create_connection_transport(bluetoothSocket, aioblescan.BLEScanRequester, None, None)
        self.connection, self.bluetoothControl = event_loop.run_until_complete(transportProcess)

        print(tfs(), "Connection made!")
        self.bluetoothControl.process = self.parsingProcess

        self.timeRequestStart = time.time()
        self.bluetoothControl.send_scan_request()
        print(tfs(), "Scan command sent!")

        alreadyCleanedUp = False

        try:
            event_loop.run_until_complete(self.awaitEventSleep(1))
            if not self.eventReceived:
                if attempt < 10:
                    print(tfs(), 'Retrying... Closing event loop', attempt)
                    self.cleanup(event_loop)
                    alreadyCleanedUp = True
                    self.scan(attempt + 1)
                    return
                else:
                    pass

            event_loop.run_until_complete(self.awaitActiveSleep(self.scanDuration))
        except KeyboardInterrupt:
            print('keyboard interrupt')
        finally:
            print("")
            if not alreadyCleanedUp:
                print(tfs(), 'closing event loop', attempt)
                self.cleanup(event_loop)


    async def awaitEventSleep(self, duration):
        while self.eventReceived == False and duration > 0:
            await asyncio.sleep(0.05)
            duration -= 0.05

    async def awaitActiveSleep(self, duration):
        while self.scanRunning == True and duration > 0:
            await asyncio.sleep(0.05)
            duration -= 0.05

    def cleanup(self, event_loop):
        print(tfs(), "Cleaning up")
        self.bluetoothControl.stop_scan_request()
        self.connection.close()
        event_loop.close()


    def parsingProcess(self, data):
        ev=aioblescan.HCI_Event()
        xx=ev.decode(data)

        hasAdvertisement = self.dataParser(ev)
        if hasAdvertisement and self.delegate is not None:
            self.delegate.handleDiscovery(ev)

    def dataParser(self, data):
        #parse Data required for the scanner
        advertisementReceived = False
        for d in data.payload:
            if isinstance(d, aioblescan.aioblescan.HCI_CC_Event):
                self.checkHCI_CC_EVENT(d)
            elif isinstance(d, aioblescan.Adv_Data):
                advertisementReceived = self.dataParser(d) or advertisementReceived
            elif isinstance(d, aioblescan.HCI_LE_Meta_Event):
                advertisementReceived = self.dataParser(d) or advertisementReceived
            elif isinstance(d, aioblescan.aioblescan.HCI_LEM_Adv_Report):
                self.eventReceived = True
                advertisementReceived = True
        return advertisementReceived


    def checkHCI_CC_EVENT(self, event):
        for d in event.payload:
            if isinstance(d, aioblescan.aioblescan.OgfOcf):
                if d.ocf == b'\x0b':
                    print(tfs(),"Settings received")
                elif d.ocf == b'\x0c':
                    print(tfs(), "Scan command received")
            # if isinstance(d, aioblescan.aioblescan.Itself):
            #     print("byte", d.name)
            # if isinstance(d, aioblescan.aioblescan.UIntByte):
            #     print("UIntByte", d.val)

    def parseAdvertisement(self, decodedHciEvent):
        global counter
        if counter % 50 == 0:
            counter = 0
            print(".")
        else:
            sys.stdout.write(".")
        counter+= 1
        # decodedHciEvent.show()



"""
Simple BLE forever-scan example, that prints all the detected
LE advertisement packets, and prints a colored diff of data on data changes.
"""
import sys
import struct
import bluetooth._bluetooth as bluez

from bluetooth_utils import (toggle_device,
                             enable_le_scan, parse_le_advertising_events,
                             disable_le_scan, raw_packet_to_str)

dev_id = 1  # the bluetooth device is hci0
toggle_device(dev_id, True)

try:
    sock = bluez.hci_open_dev(dev_id)
except:
    print("Cannot open bluetooth device %i" % dev_id)
    raise

enable_le_scan(sock, filter_duplicates=False)

import pynput.keyboard
keyboard = pynput.keyboard.Controller()
KEY_VOLUME_DOWN = pynput.keyboard.KeyCode.from_vk(0x1008ff11)
KEY_VOLUME_UP = pynput.keyboard.KeyCode.from_vk(0x1008ff13)

try:
    prev_data = None
    prev_rot_val = 0

    def le_advertise_packet_handler(mac, data, rssi):
        global prev_data
        global prev_rot_val
        print()
        print("packet len is", len(data))
        data_str = raw_packet_to_str(data)
        data_wo_rssi = (mac, data_str)
        print("BLE packet: %s %s %d" % (mac, data_str, rssi))
        if prev_data is not None:
            if data_wo_rssi != prev_data:
                # color differences with previous packet data
                sys.stdout.write(' ' * 20 + 'data_diff=')
                for c1, c2 in zip(data_str, prev_data[1]):
                    if c1 != c2:
                        sys.stdout.write('\033[0;33m' + c1 + '\033[m')
                    else:
                        sys.stdout.write(c1)
                sys.stdout.write('\n')


                # Advertisement are split into:
                # <sub-packet length byte> <sub-packet type byte> <sub-packet data (length-1)>"
                # (types are listed here: https://www.bluetooth.com/specifications/assigned-numbers/generic-access-profile)

                pkt_start = 4
                pkt_data_start = pkt_start + 2
                pkt_data_len = data[pkt_start] - 1
                assert data[pkt_start + 1] == 0x09  # Type should be "Complete Local Name"
                print("name is %r" % data[pkt_data_start:pkt_data_start + pkt_data_len].decode('utf-8'))

                pkt_start = pkt_data_start + pkt_data_len
                pkt_data_start = pkt_start + 2
                pkt_data_len = data[pkt_start] - 1
                assert data[pkt_start + 1] == 0xFF  # Type should be "Manufacturer Specific Data"
                print("manufacturer data len is", pkt_data_len)
                my_data = data[pkt_data_start:pkt_data_start + pkt_data_len]
                my_data = struct.unpack('B' * pkt_data_len, my_data)
                print("manufacturer data is", my_data)
                if my_data[2] == 1:
                    key = pynput.keyboard.Key.right
                    key = KEY_VOLUME_UP
                    keyboard.press(key); keyboard.release(key)
                elif my_data[2] == 2:
                    key = pynput.keyboard.Key.left
                    key = KEY_VOLUME_DOWN
                    keyboard.press(key); keyboard.release(key)
                prev_rot_val = my_data[1]

        prev_data = data_wo_rssi

    # Blocking call (the given handler will be called each time a new LE
    # advertisement packet is detected)
    parse_le_advertising_events(sock,
                                # mac_addr=['24:0A:C4:00:A9:72'], # wipy2
                                mac_addr=['E1:1B:2B:AC:1F:F9'],   # ble nano v2
                                handler=le_advertise_packet_handler,
                                debug=False)
except KeyboardInterrupt:
    disable_le_scan(sock)



from bluepy.btle import Scanner, Peripheral, ADDR_TYPE_RANDOM, BTLEException


from threading import Timer

from crownstone_core.util.EncryptionHandler import EncryptionHandler
from crownstone_core.Exceptions import CrownstoneBleException
from crownstone_core.protocol.BluenetTypes import ProcessType

from crownstone_ble.Exceptions import BleError
from crownstone_ble.constants import ScanBackends
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.core.bluetooth_delegates.AioScanDelegate import AioScanDelegate
from crownstone_ble.core.bluetooth_delegates.AioScanner import AioScanner
from crownstone_ble.core.bluetooth_delegates.ScanDelegate import ScanDelegate
from crownstone_ble.core.bluetooth_delegates.SingleNotificationDelegate import PeripheralDelegate
from crownstone_ble.core.modules.Validator import Validator
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics

CCCD_UUID = 0x2902

class BleHandler:
    
    def __init__(self, settings, hciIndex=0, scanBackend = ScanBackends.Bluepy):
        self.connectedPeripherals = {}

        self.settings = None

        self.connectedPeripherals = {}
        self.connectedPeripheral = None

        self.notificationLoopActive = False
        self.notificationResult = None

        self.scanningActive = False
        self.scanAborted = False

        self.subscriptionIds = []
        
        self.validator = Validator()
        self.settings = settings
        self.hciIndex = hciIndex
        self.scanBackend = scanBackend

        if self.scanBackend == ScanBackends.Aio:
            self.scanner = AioScanner(self.hciIndex).withDelegate(AioScanDelegate(settings))
        else:
            self.scanner = Scanner(self.hciIndex).withDelegate(ScanDelegate(settings))
        self.subscriptionIds.append(BleEventBus.subscribe(SystemBleTopics.abortScanning, lambda x: self.abortScanning()))
        
    
    def shutDown(self):
        for subscriptionId in self.subscriptionIds:
            BleEventBus.unsubscribe(subscriptionId)
            
        self.validator.shutDown()
    
    
    def connect(self, address, connectionSettings = None):
        if address not in self.connectedPeripherals:
            self.connectedPeripherals[address] = Peripheral(iface=self.hciIndex)
            print("Connecting...")
            self.connectedPeripheral = address
            self.connectedPeripherals[address].connect(address, addrType=ADDR_TYPE_RANDOM, iface=self.hciIndex)
            self.connectedPeripherals[address].getServices()
            print("Connected")
            if connectionSettings is not None:
                if connectionSettings.mtu:
                    print("Set MTU")
                    self.connectedPeripherals[address].setMTU(connectionSettings.mtu)
            
    
    def disconnect(self):
        print("Disconnecting... Cleaning up")
        if self.connectedPeripheral:
            self.connectedPeripherals[self.connectedPeripheral].disconnect()
            del self.connectedPeripherals[self.connectedPeripheral]
            self.connectedPeripheral = None
            print("Cleaned up")
    
    
    def startScanning(self, scanDuration=3):
        if not self.scanningActive:
            self.scanningActive = True
            self.scanAborted = False
            if self.scanBackend == ScanBackends.Aio:
                self.scanner.start(scanDuration)
            else:
                self.scanner.start()
                scanTime = 0
                processInterval = 0.5
                while self.scanningActive and scanTime < scanDuration and not self.scanAborted:
                    scanTime += processInterval
                    self.scanner.process(processInterval)

                self.stopScanning()

    def startScanningBackground(self, scanDuration=3):
        Timer(0.0001, lambda: self.startScanning(scanDuration))

    
    def stopScanning(self):
        if self.scanningActive:
            self.scanner.stop()
            self.scanningActive = False
            
    def abortScanning(self):
        if self.scanningActive:
            self.scanAborted = True
            if self.scanBackend == ScanBackends.Aio:
                self.scanner.stop()
    
    def enableNotifications(self):
        print("ENABLE NOTIFICATIONS IS NOT IMPLEMENTED YET")
    
    def disableNotifications(self):
        print("DISABLE NOTIFICATIONS IS NOT IMPLEMENTED YET")

    def writeToCharacteristic(self, serviceUUID, characteristicUUID, content):
        targetCharacteristic = self.getCharacteristic(serviceUUID, characteristicUUID)
        encryptedContent = EncryptionHandler.encrypt(content, self.settings)
        targetCharacteristic.write(encryptedContent, withResponse=True)

    def writeToCharacteristicWithoutEncryption(self, serviceUUID, characteristicUUID, content):
        byteContent = bytes(content)
        targetCharacteristic = self.getCharacteristic(serviceUUID, characteristicUUID)
        targetCharacteristic.write(byteContent, withResponse=True)

    def readCharacteristic(self, serviceUUID, characteristicUUID):
        data = self.readCharacteristicWithoutEncryption(serviceUUID, characteristicUUID)
        if self.settings.isEncryptionEnabled():
            return EncryptionHandler.decrypt(data, self.settings)


    def readCharacteristicWithoutEncryption(self, serviceUUID, characteristicUUID):
        targetCharacteristic = self.getCharacteristic(serviceUUID, characteristicUUID)
        data = targetCharacteristic.read()
        return data



    def getCharacteristics(self, serviceUUID):
        if self.connectedPeripheral:
            peripheral = self.connectedPeripherals[self.connectedPeripheral]

            try:
                service = peripheral.getServiceByUUID(serviceUUID)
            except BTLEException:
                raise CrownstoneBleException(BleError.CAN_NOT_FIND_SERVICE, "Can not find service: " + serviceUUID)

            characteristics = service.getCharacteristics()

            return characteristics

        else:
            raise CrownstoneBleException(BleError.CAN_NOT_GET_CHACTERISTIC, "Can't get characteristics: Not connected.")


    def getCharacteristic(self, serviceUUID, characteristicUUID):
        if self.connectedPeripheral:
            peripheral = self.connectedPeripherals[self.connectedPeripheral]
        
            try:
                service = peripheral.getServiceByUUID(serviceUUID)
            except BTLEException:
                raise CrownstoneBleException(BleError.CAN_NOT_FIND_SERVICE, "Can not find service: " + serviceUUID)
        
            characteristics = service.getCharacteristics(characteristicUUID)
            if len(characteristics) == 0:
                raise CrownstoneBleException(BleError.CAN_NOT_FIND_CHACTERISTIC, "Can not find characteristic: " + characteristicUUID)

            return characteristics[0]
        
        else:
            raise CrownstoneBleException(BleError.CAN_NOT_GET_CHACTERISTIC, "Can't get characteristic: Not connected.")
        
        
    def setupSingleNotification(self, serviceUUID, characteristicUUID, writeCommand):
        characteristic = self.getCharacteristic(serviceUUID, characteristicUUID)
        peripheral = self.connectedPeripherals[self.connectedPeripheral]
        
        peripheral.withDelegate(PeripheralDelegate(lambda x: self._killNotificationLoop(x), self.settings))
        
        characteristicCCCDList = characteristic.getDescriptors(forUUID=CCCD_UUID)
        if len(characteristicCCCDList) == 0:
            raise CrownstoneBleException(BleError.CAN_NOT_FIND_CCCD, "Can not find CCCD handle to use notifications for characteristic: " + characteristicUUID)
        
        characteristicCCCD = characteristicCCCDList[0]
        
        # enable notifications.. This is ugly but necessary
        characteristicCCCD.write(b"\x01\x00", True)
        
        # execute something that will trigger the notifications
        writeCommand()
        
        self.notificationLoopActive = True

        loopCount = 0
        while self.notificationLoopActive and loopCount < 10:
            peripheral.waitForNotifications(0.5)
            loopCount += 1


        if self.notificationResult is None:
            raise CrownstoneBleException(BleError.NO_NOTIFICATION_DATA_RECEIVED, "No notification data received.")
        
        result = self.notificationResult
        self.notificationResult = None
        
        return result
    
    def setupNotificationStream(self, serviceUUID, characteristicUUID, writeCommand, resultHandler, timeout):
        characteristic = self.getCharacteristic(serviceUUID, characteristicUUID)
        peripheral = self.connectedPeripherals[self.connectedPeripheral]
        
        peripheral.withDelegate(PeripheralDelegate(lambda x: self._loadNotificationResult(x), self.settings))
    
        characteristicCCCDList = characteristic.getDescriptors(forUUID=CCCD_UUID)
        if len(characteristicCCCDList) == 0:
            raise CrownstoneBleException(BleError.CAN_NOT_FIND_CCCD, "Can not find CCCD handle to use notifications for characteristic: " + characteristicUUID)
    
        characteristicCCCD = characteristicCCCDList[0]
    
        # enable notifications.. This is ugly but necessary
        characteristicCCCD.write(b"\x01\x00", True)
    
        # execute something that will trigger the notifications
        writeCommand()
    
        self.notificationLoopActive = True
        self.notificationResult = None
        
        loopCount = 0
        successful = False
        while self.notificationLoopActive and loopCount < timeout*2:
            peripheral.waitForNotifications(0.5)
            loopCount += 1
            if self.notificationResult is not None:
                command = resultHandler(self.notificationResult)
                self.notificationResult = None
                if command == ProcessType.ABORT_ERROR:
                    self.notificationLoopActive = False
                    raise CrownstoneBleException(BleError.ABORT_NOTIFICATION_STREAM_W_ERROR, "Aborting the notification stream because the resultHandler raised an error.")
                elif command == ProcessType.FINISHED:
                    self.notificationLoopActive = False
                    successful = True
    
        if not successful:
            raise CrownstoneBleException(BleError.NOTIFICATION_STREAM_TIMEOUT, "Notification stream not finished within timeout.")
    
        
    def _killNotificationLoop(self, result):
        self.notificationLoopActive = False
        self.notificationResult = result
        
    def _loadNotificationResult(self, result):
        self.notificationResult = result




from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics


class NearestSelector:
    
    def __init__(self, setupModeOnly=False, rssiAtLeast=-100, returnFirstAcceptable=False, addressesToExcludeSet=set()):
        self.setupModeOnly = setupModeOnly
        self.rssiAtLeast = rssiAtLeast
        self.returnFirstAcceptable = returnFirstAcceptable
        self.addressesToExcludeSet = addressesToExcludeSet
        self.deviceList = []
        self.nearest = None
        
        
    def handleAdvertisement(self, advertisement):
        if "serviceData" not in advertisement:
            return
        
        if advertisement["address"].lower() in self.addressesToExcludeSet:
            return
        
        if self.setupModeOnly and not advertisement["serviceData"]["setupMode"]:
            return
        
        
        if not self.setupModeOnly and advertisement["serviceData"]["setupMode"]:
            return
            
            
        if advertisement["rssi"] < self.rssiAtLeast:
            return
        
        self.deviceList.append(advertisement)
        
        if self.returnFirstAcceptable:
            BleEventBus.emit(SystemBleTopics.abortScanning, True)
            
            
    def getNearest(self):
        
        if len(self.deviceList) == 0:
            return None
        
        nearest = self.deviceList[0]
        
        for adv in self.deviceList:
            if nearest["rssi"] < adv["rssi"] < 0:
                nearest = adv
            
        return {
            "name": nearest["name"],
            "address": nearest["address"],
            "rssi" : nearest["rssi"],
            "setupMode": nearest["serviceData"]["setupMode"],
            "id" : nearest["serviceData"]["id"]
        }
        
        
        
        


# -*- coding: utf-8 -*-
"""
Module containing some bluetooth utility functions (linux only).

It either uses HCI commands using PyBluez, or does ioctl calls like it's
done in Bluez tools such as hciconfig.

Main functions:
  - toggle_device : enable or disable a bluetooth device
  - set_scan : set scan type on a device ("noscan", "iscan", "pscan", "piscan")
  - enable/disable_le_scan : enable BLE scanning
  - parse_le_advertising_events : parse and read BLE advertisements packets
  - start/stop_le_advertising : advertise custom data using BLE

Bluez : http://www.bluez.org/
PyBluez : http://karulis.github.io/pybluez/

The module was in particular inspired from 'iBeacon-Scanner-'
https://github.com/switchdoclabs/iBeacon-Scanner-/blob/master/blescan.py
and sometimes directly from the Bluez sources.
"""

from __future__ import absolute_import
import sys
import struct
import fcntl
import array
import socket
from errno import EALREADY

# import PyBluez
import bluetooth._bluetooth as bluez

__all__ = ('toggle_device', 'set_scan',
           'enable_le_scan', 'disable_le_scan', 'parse_le_advertising_events',
           'start_le_advertising', 'stop_le_advertising',
           'raw_packet_to_str')

LE_META_EVENT = 0x3E
LE_PUBLIC_ADDRESS = 0x00
LE_RANDOM_ADDRESS = 0x01

OGF_LE_CTL = 0x08
OCF_LE_SET_SCAN_PARAMETERS = 0x000B
OCF_LE_SET_SCAN_ENABLE = 0x000C
OCF_LE_CREATE_CONN = 0x000D
OCF_LE_SET_ADVERTISING_PARAMETERS = 0x0006
OCF_LE_SET_ADVERTISE_ENABLE = 0x000A
OCF_LE_SET_ADVERTISING_DATA = 0x0008

SCAN_TYPE_PASSIVE = 0x00
SCAN_FILTER_DUPLICATES = 0x01
SCAN_DISABLE = 0x00
SCAN_ENABLE = 0x01

# sub-events of LE_META_EVENT
EVT_LE_CONN_COMPLETE = 0x01
EVT_LE_ADVERTISING_REPORT = 0x02
EVT_LE_CONN_UPDATE_COMPLETE = 0x03
EVT_LE_READ_REMOTE_USED_FEATURES_COMPLETE = 0x04

# Advertisement event types
ADV_IND = 0x00
ADV_DIRECT_IND = 0x01
ADV_SCAN_IND = 0x02
ADV_NONCONN_IND = 0x03
ADV_SCAN_RSP = 0x04

# Allow Scan Request from Any, Connect Request from Any
FILTER_POLICY_NO_WHITELIST = 0x00
# Allow Scan Request from White List Only, Connect Request from Any
FILTER_POLICY_SCAN_WHITELIST = 0x01
# Allow Scan Request from Any, Connect Request from White List Only
FILTER_POLICY_CONN_WHITELIST = 0x02
# Allow Scan Request from White List Only, Connect Request from White List Only
FILTER_POLICY_SCAN_AND_CONN_WHITELIST = 0x03


def toggle_device(dev_id, enable):
    """
    Power ON or OFF a bluetooth device.

    :param dev_id: Device id.
    :type dev_id: ``int``
    :param enable: Whether to enable of disable the device.
    :type enable: ``bool``
    """
    hci_sock = socket.socket(socket.AF_BLUETOOTH,
                             socket.SOCK_RAW,
                             socket.BTPROTO_HCI)
    print("Power %s bluetooth device %d" % ('ON' if enable else 'OFF', dev_id))
    # di = struct.pack("HbBIBBIIIHHHH10I", dev_id, *((0,) * 22))
    # fcntl.ioctl(hci_sock.fileno(), bluez.HCIGETDEVINFO, di)
    req_str = struct.pack("H", dev_id)
    request = array.array("b", req_str)
    try:
        fcntl.ioctl(hci_sock.fileno(),
                    bluez.HCIDEVUP if enable else bluez.HCIDEVDOWN,
                    request[0])
    except IOError as e:
        if e.errno == EALREADY:
            print("Bluetooth device %d is already %s" % (
                  dev_id, 'enabled' if enable else 'disabled'))
        else:
            raise
    finally:
        hci_sock.close()


# Types of bluetooth scan
SCAN_DISABLED = 0x00
SCAN_INQUIRY = 0x01
SCAN_PAGE = 0x02


def set_scan(dev_id, scan_type):
    """
    Set scan type on a given bluetooth device.

    :param dev_id: Device id.
    :type dev_id: ``int``
    :param scan_type: One of
        ``'noscan'``
        ``'iscan'``
        ``'pscan'``
        ``'piscan'``
    :type scan_type: ``str``
    """
    hci_sock = socket.socket(socket.AF_BLUETOOTH,
                             socket.SOCK_RAW,
                             socket.BTPROTO_HCI)
    if scan_type == "noscan":
        dev_opt = SCAN_DISABLED
    elif scan_type == "iscan":
        dev_opt = SCAN_INQUIRY
    elif scan_type == "pscan":
        dev_opt = SCAN_PAGE
    elif scan_type == "piscan":
        dev_opt = SCAN_PAGE | SCAN_INQUIRY
    else:
        raise ValueError("Unknown scan type %r" % scan_type)

    req_str = struct.pack("HI", dev_id, dev_opt)
    print("Set scan type %r to bluetooth device %d" % (scan_type, dev_id))
    try:
        fcntl.ioctl(hci_sock.fileno(), bluez.HCISETSCAN, req_str)
    finally:
        hci_sock.close()


def raw_packet_to_str(pkt):
    """
    Returns the string representation of a raw HCI packet.
    """
    if sys.version_info > (3, 0):
        return ''.join('%02x' % struct.unpack("B", bytes([x]))[0] for x in pkt)
    else:
        return ''.join('%02x' % struct.unpack("B", x)[0] for x in pkt)


def enable_le_scan(sock, interval=0x0800, window=0x0800,
                   filter_policy=FILTER_POLICY_NO_WHITELIST,
                   filter_duplicates=True):
    """
    Enable LE passive scan (with filtering of duplicate packets enabled).

    :param sock: A bluetooth HCI socket (retrieved using the
        ``hci_open_dev`` PyBluez function).
    :param interval: Scan interval.
    :param window: Scan window (must be less or equal than given interval).
    :param filter_policy: One of
        ``FILTER_POLICY_NO_WHITELIST`` (default value)
        ``FILTER_POLICY_SCAN_WHITELIST``
        ``FILTER_POLICY_CONN_WHITELIST``
        ``FILTER_POLICY_SCAN_AND_CONN_WHITELIST``

    .. note:: Scan interval and window are to multiply by 0.625 ms to
        get the real time duration.
    """
    print("Enable LE scan")
    own_bdaddr_type = LE_PUBLIC_ADDRESS  # does not work with LE_RANDOM_ADDRESS
    cmd_pkt = struct.pack("<BHHBB", SCAN_TYPE_PASSIVE, interval, window,
                          own_bdaddr_type, filter_policy)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_SCAN_PARAMETERS, cmd_pkt)
    print("scan params: interval=%.3fms window=%.3fms own_bdaddr=%s "
          "whitelist=%s" %
          (interval * 0.625, window * 0.625,
           'public' if own_bdaddr_type == LE_PUBLIC_ADDRESS else 'random',
           'yes' if filter_policy in (FILTER_POLICY_SCAN_WHITELIST,
                                      FILTER_POLICY_SCAN_AND_CONN_WHITELIST)
           else 'no'))
    cmd_pkt = struct.pack("<BB", SCAN_ENABLE, SCAN_FILTER_DUPLICATES if filter_duplicates else 0x00)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_SCAN_ENABLE, cmd_pkt)


def disable_le_scan(sock):
    """
    Disable LE scan.

    :param sock: A bluetooth HCI socket (retrieved using the
        ``hci_open_dev`` PyBluez function).
    """
    print("Disable LE scan")
    cmd_pkt = struct.pack("<BB", SCAN_DISABLE, 0x00)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_SCAN_ENABLE, cmd_pkt)


def start_le_advertising(sock, min_interval=1000, max_interval=1000, adv_type=ADV_NONCONN_IND, data=()):
    """
    Start LE advertising.

    :param sock: A bluetooth HCI socket (retrieved using the
        ``hci_open_dev`` PyBluez function).
    :param min_interval: Minimum advertising interval.
    :param max_interval: Maximum advertising interval.
    :param adv_type: Advertisement type (``ADV_NONCONN_IND`` by default).
    :param data: The advertisement data (maximum of 31 bytes).
    :type data: iterable
    """
    own_bdaddr_type = 0
    direct_bdaddr_type = 0
    direct_bdaddr = (0,) * 6
    chan_map = 0x07  # All channels: 37, 38, 39
    filter = 0

    struct_params = [min_interval, max_interval, adv_type, own_bdaddr_type,
                     direct_bdaddr_type]
    struct_params.extend(direct_bdaddr)
    struct_params.extend((chan_map, filter))

    cmd_pkt = struct.pack("<HHBBB6BBB", *struct_params)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_ADVERTISING_PARAMETERS,
                       cmd_pkt)

    cmd_pkt = struct.pack("<B", 0x01)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_ADVERTISE_ENABLE, cmd_pkt)

    data_length = len(data)
    if data_length > 31:
        raise ValueError("data is too long (%d but max is 31 bytes)",
                         data_length)
    cmd_pkt = struct.pack("<B%dB" % data_length, data_length, *data)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_ADVERTISING_DATA, cmd_pkt)
    print("Advertising started data_length=%d data=%r" % (data_length, data))


def stop_le_advertising(sock):
    """
    Stop LE advertising.

    :param sock: A bluetooth HCI socket (retrieved using the
        ``hci_open_dev`` PyBluez function).
    """
    cmd_pkt = struct.pack("<B", 0x00)
    bluez.hci_send_cmd(sock, OGF_LE_CTL, OCF_LE_SET_ADVERTISE_ENABLE, cmd_pkt)
    print("Advertising stopped")


def parse_le_advertising_events(sock, mac_addr=None, packet_length=None,
                                handler=None, debug=False):
    """
    Parse and report LE advertisements.

    This is a blocking call, an infinite loop is started and the
    given handler will be called each time a new LE advertisement packet
    is detected and corresponds to the given filters.

    .. note:: The :func:`.start_le_advertising` function must be
        called before calling this function.

    :param sock: A bluetooth HCI socket (retrieved using the
        ``hci_open_dev`` PyBluez function).
    :param mac_addr: list of filtered mac address representations
        (uppercase, with ':' separators).
        If not specified, the LE advertisement of any device will be reported.
        Example: mac_addr=('00:2A:5F:FF:25:11', 'DA:FF:12:33:66:12')
    :type mac_addr: ``list`` of ``string``
    :param packet_length: Filter a specific length of LE advertisement packet.
    :type packet_length: ``int``
    :param handler: Handler that will be called each time a LE advertisement
        packet is available (in accordance with the ``mac_addr``
        and ``packet_length`` filters).
    :type handler: ``callable`` taking 4 parameters:
        mac (``str``), adv_type (``int``), data (``bytes``) and rssi (``int``)
    :param debug: Enable debug prints.
    :type debug: ``bool``
    """
    if not debug and handler is None:
        raise ValueError("You must either enable debug or give a handler !")

    old_filter = sock.getsockopt(bluez.SOL_HCI, bluez.HCI_FILTER, 14)

    flt = bluez.hci_filter_new()
    bluez.hci_filter_set_ptype(flt, bluez.HCI_EVENT_PKT)
    # bluez.hci_filter_all_events(flt)
    bluez.hci_filter_set_event(flt, LE_META_EVENT)
    sock.setsockopt(bluez.SOL_HCI, bluez.HCI_FILTER, flt)

    print("socket filter set to ptype=HCI_EVENT_PKT event=LE_META_EVENT")
    print("Listening ...")

    try:
        while True:
            pkt = sock.recv(255)
            ptype, event, plen = struct.unpack("BBB", pkt[:3])

            if event != LE_META_EVENT:
                # Should never occur because we filtered with this type of event
                print("Not a LE_META_EVENT !")
                continue

            sub_event, = struct.unpack("B", pkt[3:4])
            if sub_event != EVT_LE_ADVERTISING_REPORT:
                if debug:
                    print("Not a EVT_LE_ADVERTISING_REPORT !")
                continue

            pkt = pkt[4:]
            adv_type = struct.unpack("b", pkt[1:2])[0]
            mac_addr_str = bluez.ba2str(pkt[3:9])

            if packet_length and plen != packet_length:
                # ignore this packet
                if debug:
                    print("packet with non-matching length: mac=%s adv_type=%02x plen=%s" %
                          (mac_addr_str, adv_type, plen))
                    print(raw_packet_to_str(pkt))
                continue

            data = pkt[9:-1]
            rssi = struct.unpack("b", pkt[-2:-1])[0]

            if mac_addr and mac_addr_str not in mac_addr:
                if debug:
                    print("packet with non-matching mac %s adv_type=%02x data=%s RSSI=%s" %
                          (mac_addr_str, adv_type, raw_packet_to_str(data), rssi))
                continue

            if debug:
                print("LE advertisement: mac=%s adv_type=%02x data=%s RSSI=%d" %
                      (mac_addr_str, adv_type, raw_packet_to_str(data), rssi))

            if handler is not None:
                try:
                    handler(mac_addr_str, adv_type, data, rssi)
                except Exception as e:
                    print('Exception when calling handler with a BLE advertising event: %r' % (e,))
                    import traceback
                    traceback.print_exc()

    except KeyboardInterrupt:
        print("\nRestore previous socket filter")
        sock.setsockopt(bluez.SOL_HCI, bluez.HCI_FILTER, old_filter)
        raise

"""
def hci_le_add_white_list(int dd, const bdaddr_t *bdaddr, uint8_t type, int to)
{
    struct hci_request {
        uint16_t ogf;
        uint16_t ocf;
        int      event;
        void     *cparam;
        int      clen;
        void     *rparam;
        int      rlen;
    };

    struct hci_request rq;
    le_add_device_to_white_list_cp cp;
    uint8_t status;

    memset(&cp, 0, sizeof(cp));
    cp.bdaddr_type = type;
    bacpy(&cp.bdaddr, bdaddr);

    memset(&rq, 0, sizeof(rq));
    rq.ogf = OGF_LE_CTL;
    rq.ocf = OCF_LE_ADD_DEVICE_TO_WHITE_LIST;
    rq.cparam = &cp;
    rq.clen = LE_ADD_DEVICE_TO_WHITE_LIST_CP_SIZE;
    rq.rparam = &status;
    rq.rlen = 1;

    if (hci_send_req(dd, &rq, to) < 0)
            return -1;

    if (status) {
            errno = EIO;
            return -1;
    }

    return 0;
}"""



from bluepy.btle import DefaultDelegate
from crownstone_core.Exceptions import CrownstoneBleException
from crownstone_core.util.EncryptionHandler import EncryptionHandler

LAST_PACKET_INDEX = 0xFF

class PeripheralDelegate(DefaultDelegate):
    
    def __init__(self, callback, settings):
        DefaultDelegate.__init__(self)
        self.callback = callback
        self.dataCollected = []
        self.settings = settings

    def handleNotification(self, cHandle, data):
        self.merge(data)
        
    def merge(self, data):
        self.dataCollected += data[1:]
        
        if data[0] == LAST_PACKET_INDEX:
            data = self.checkPayload()
            self.dataCollected = []
            self.callback(data)
                
       
    def checkPayload(self):
        try:
            return EncryptionHandler.decrypt(self.dataCollected, self.settings)
        except CrownstoneBleException as err:
            print(err)
        


import time
from bluepy.btle import BTLEException
from crownstone_core.packets.MicroappPacket import MicroappPacketInternal
from crownstone_core.protocol.BluenetTypes import ProcessType

from crownstone_ble.Exceptions import BleError
from crownstone_core.Exceptions import CrownstoneException, CrownstoneBleException
from crownstone_core.packets.SessionDataPacket import SessionDataPacket
from crownstone_core.protocol.Characteristics import CrownstoneCharacteristics
from crownstone_core.protocol.ControlPackets import ControlPacketsGenerator
from crownstone_core.protocol.Services import CSServices
from crownstone_core.util.EncryptionHandler import EncryptionHandler, CHECKSUM


class ControlHandler:
    def __init__(self, bluetoothCore):
        self.core = bluetoothCore

    def getAndSetSessionNone(self):
        # read the nonce
        rawNonce = self.core.ble.readCharacteristicWithoutEncryption(CSServices.CrownstoneService, CrownstoneCharacteristics.SessionData)
        ProcessSessionNoncePacket(rawNonce, self.core.settings.basicKey, self.core.settings)

    def setSwitchState(self, switchState):
        """
         :param switchState: number [0..1]
        """
        self._writeControlPacket(ControlPacketsGenerator.getSwitchStatePacket(switchState))

    def switchRelay(self, switchState):
        """
         :param switchState: number 0 is off, 1 is on.
        """
        self._writeControlPacket(ControlPacketsGenerator.getRelaySwitchPacket(switchState))

    def switchPWM(self, switchState):
        """
         :param switchState: number [0..1]
        """
        self._writeControlPacket(ControlPacketsGenerator.getPwmSwitchPacket(switchState))

    def commandFactoryReset(self):
        """
          If you have the keys, you can use this to put the crownstone back into factory default mode
        """
        self._writeControlPacket(ControlPacketsGenerator.getCommandFactoryResetPacket())

    def allowDimming(self, allow):
        """
        :param allow: bool
        """
        self._writeControlPacket(ControlPacketsGenerator.getAllowDimmingPacket(allow))

    def disconnect(self):
        """
          This forces the Crownstone to disconnect from you
        """
        try:
            self._writeControlPacket(ControlPacketsGenerator.getDisconnectPacket())
            self.core.ble.disconnect()
        except BTLEException as err:
            if err.code is BTLEException.DISCONNECTED:
                pass
            else:
                raise err



    def lockSwitch(self, lock):
        """
        :param lock: bool
        """
        self._writeControlPacket(ControlPacketsGenerator.getLockSwitchPacket(lock))


    def reset(self):
        self._writeControlPacket(ControlPacketsGenerator.getResetPacket())



    def recovery(self, address):
        self.core.connect(address, ignoreEncryption=True)
        self._recoveryByFactoryReset()
        self._checkRecoveryProcess()
        self.core.disconnect()
        time.sleep(5)
        self.core.connect(address, ignoreEncryption=True)
        self._recoveryByFactoryReset()
        self._checkRecoveryProcess()
        self.core.disconnect()
        time.sleep(2)

    def _recoveryByFactoryReset(self):
        packet = ControlPacketsGenerator.getFactoryResetPacket()
        return self.core.ble.writeToCharacteristicWithoutEncryption(
            CSServices.CrownstoneService,
            CrownstoneCharacteristics.FactoryReset,
            packet
        )

    def _checkRecoveryProcess(self):
        result = self.core.ble.readCharacteristicWithoutEncryption(CSServices.CrownstoneService, CrownstoneCharacteristics.FactoryReset)
        if result[0] == 1:
            return True
        elif result[0] == 2:
            raise CrownstoneException(BleError.RECOVERY_MODE_DISABLED, "The recovery mechanism has been disabled by the Crownstone owner.")
        else:
            raise CrownstoneException(BleError.NOT_IN_RECOVERY_MODE, "The recovery mechanism has expired. It is only available briefly after the Crownstone is powered on.")

    def sendMicroapp(self, data):
        """
        :param data: byte array
        """
        self._microapp = MicroappPacketInternal(data)
        self.sendMicroappInternal(True)

    def sendMicroappInternal(self, firstTime, notificationResult=None):
        if not firstTime:

            if not self._microapp.nextAvailable():
                print("LOG: Finish stream")
                return ProcessType.FINISHED

            self._microapp.update()

        # logStr = "LOG: write part %i of %i (size = %i)\n" % (self._microapp.index + 1, self._microapp.count, self._microapp.data.size)
        # sys.stdout.write(logStr);

        if firstTime:
            timeout = self._microapp.count * 10
            print("LOG: Use (for the overall connection) a timeout of", timeout)

            self.core.ble.setupNotificationStream(
                CSServices.CrownstoneService,
                CrownstoneCharacteristics.Result,
                lambda: self._writeControlPacket(ControlPacketsGenerator.getMicroAppPacket(
                    self._microapp.getPacket())),
                lambda notificationResult: self._handleResult(notificationResult),
                timeout
            )
        else:
            self._writeControlPacket(ControlPacketsGenerator.getMicroAppPacket(
                    self._microapp.getPacket()))
            return ProcessType.CONTINUE

    def enableMicroapp(self, offset):
        """
        :param data: byte array
        """
        self._writeControlPacket(ControlPacketsGenerator.getMicroAppMetaPacket(self._microapp.getMetaPacket(offset)))

    def _handleResult(self, notificationResult):
        if notificationResult:
            print(notificationResult)
            err_code = notificationResult[2]
            # Only print atypical error codes
            if err_code != 0x00 and err_code != 0x01:
                print("Err code" , err_code)

            # Display type of return code (error)
            if err_code == 0x20:
                print("LOG: ERR_WRONG_PAYLOAD_LENGTH")
            if err_code == 0x70:
                print("LOG: ERR_EVENT_UNHANDLED")
            if err_code == 0x10:
                print("LOG: NRF_ERROR_INVALID_ADDR")
            if err_code == 0x27:
                print("LOG: ERR_BUSY")
            if err_code == 0x22:
                print("LOG: ERR_INVALID_MESSAGE")

            # Normal error codes (first wait for success, then success)
            if err_code == 0x01:
                print("LOG: ERR_WAIT_FOR_SUCCESS")
            if err_code == 0x00:
                print("LOG: ERR_SUCCESS")
                return self.sendMicroappInternal(False, notificationResult)

    #  self.bleManager.readCharacteristic(CSServices.CrownstoneService, characteristicId: CrownstoneCharacteristics.FactoryReset)
    # {(result:[UInt8]) -> Void in
    # if (result[0] == 1)
    #     seal.fulfill(())
    # else if (result[0] == 2) {
    # seal.reject(CrownstoneError.RECOVER_MODE_DISABLED)
    # else {
    # seal.reject(CrownstoneError.NOT_IN_RECOVERY_MODE)
    # .catch{(err) -> Void in
    # seal.reject(CrownstoneError.CANNOT_READ_FACTORY_RESET_CHARACTERISTIC)


    # return self.bleManager.connect(uuid)}
    # .then
    # {(_) -> Promise < Void > in
    # return self._recoverByFactoryReset()}
    # .then
    # {(_) -> Promise < Void > in
    # return self._checkRecoveryProcess()}
    # .then
    # {(_) -> Promise < Void > in
    # return self.bleManager.disconnect()}
    # .then
    # {(_) -> Promise < Void > in
    # return self.bleManager.waitToReconnect()}
    # .then
    # {(_) -> Promise < Void > in
    # return self.bleManager.connect(uuid)}
    # .then
    # {(_) -> Promise < Void > in
    # return self._recoverByFactoryReset()}
    # .then
    # {(_) -> Promise < Void > in
    # return self._checkRecoveryProcess()}
    # .then
    # {(_) -> Promise < Void > in
    # self.bleManager.settings.restoreEncryption()
    # return self.bleManager.disconnect()



    """
    ---------------  UTIL  ---------------
    """




    def _readControlPacket(self, packet):
        return self.core.ble.readCharacteristic(CSServices.CrownstoneService, CrownstoneCharacteristics.Control)

    def _writeControlPacket(self, packet):
        self.core.ble.writeToCharacteristic(CSServices.CrownstoneService, CrownstoneCharacteristics.Control, packet)


def ProcessSessionNoncePacket(encryptedPacket, key, settings):
    # decrypt it
    decrypted = EncryptionHandler.decryptECB(encryptedPacket, key)

    packet = SessionDataPacket(decrypted)
    if packet.validation == CHECKSUM:
        # load into the settings object
        settings.setSessionNonce(packet.sessionNonce)
        settings.setValidationKey(packet.validationKey)
        settings.setCrownstoneProtocolVersion(packet.protocol)
    else:
        raise CrownstoneBleException(BleError.COULD_NOT_VALIDATE_SESSION_NONCE, "Could not validate the session nonce.")


from bluepy.btle import DefaultDelegate
from crownstone_core.packets.Advertisement import Advertisement

from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics

SERVICE_DATA_ADTYPE = 22
NAME_ADTYPE         = 8
FLAGS_ADTYPE        = 1

class ScanDelegate(DefaultDelegate):
    
    def __init__(self, settings):
        self.settings = settings
        DefaultDelegate.__init__(self)

    def handleDiscovery(self, dev, isNewDev, isNewData):
        valueText = dev.getValueText(SERVICE_DATA_ADTYPE)
        nameText = dev.getValueText(NAME_ADTYPE)
        if valueText is not None:
            self.parsePayload(dev.addr, dev.rssi, nameText, valueText)
          
    def parsePayload(self, address, rssi, nameText, valueText):
        advertisement = Advertisement(address, rssi, nameText, valueText)

        if advertisement.serviceData.opCode <= 5:
            advertisement.decrypt(self.settings.basicKey)
        elif advertisement.serviceData.opCode >= 7:
            advertisement.decrypt(self.settings.serviceDataKey)

        if advertisement.isCrownstoneFamily():
            BleEventBus.emit(SystemBleTopics.rawAdvertisement, advertisement)


import threading

from crownstone_core.topics.Topics import Topics

from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.core.modules.StoneAdvertisementTracker import StoneAdvertisementTracker
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics


class Validator:

    def __init__(self):
        BleEventBus.subscribe(SystemBleTopics.rawAdvertisement, self.checkAdvertisement)
        self.tickTimer = None
        self._lock = threading.Lock()
        self.scheduleTick()
        self.trackedCrownstones = {}


    def scheduleTick(self):
        if self.tickTimer is not None:
            self.tickTimer.cancel()

        self.tickTimer = threading.Timer(1, lambda: self.tick())
        self.tickTimer.start()


    def tick(self):
        with self._lock:
            allKeys = []
            # we first collect keys because we might delete items from this list during ticks
            for key, trackedStone in self.trackedCrownstones.items():
                allKeys.append(key)

            for key in allKeys:
                self.trackedCrownstones[key].tick()

        self.scheduleTick()


    def removeStone(self, address):
        del self.trackedCrownstones[address]


    def checkAdvertisement(self, advertisement):
        if advertisement.address not in self.trackedCrownstones:
            self.trackedCrownstones[advertisement.address] = StoneAdvertisementTracker(lambda: self.removeStone(advertisement.address))
        
        self.trackedCrownstones[advertisement.address].update(advertisement)
        
        if self.trackedCrownstones[advertisement.address].verified:
            BleEventBus.emit(Topics.advertisement, advertisement.getDictionary())
            
            if advertisement.hasScanResponse:
                BleEventBus.emit(Topics.newDataAvailable, advertisement.getSummary())


    def shutDown(self):
        if self.tickTimer is not None:
            self.tickTimer.cancel()



from crownstone_core import Conversion
from crownstone_core.Exceptions import CrownstoneError, CrownstoneException
from crownstone_core.packets.ResultPacket import ResultPacket
from crownstone_core.protocol.BlePackets import ControlStateGetPacket
from crownstone_core.protocol.BluenetTypes import StateType
from crownstone_core.protocol.Characteristics import CrownstoneCharacteristics
from crownstone_core.protocol.Services import CSServices


class StateHandler:
    def __init__(self, bluetoothCore):
        self.core = bluetoothCore
        
    def getSwitchState(self):
        return self._getState(StateType.SWITCH_STATE)[0]
    
    def getSwitchStateFloat(self):
        switchState = self._getState(StateType.SWITCH_STATE)[0]
        
        returnState = 0.0
        if switchState == 128:
            returnState = 1.0
        elif switchState <= 100:
            returnState = 0.01 * switchState * 0.99
        
        return returnState
    
    def getTime(self):
        bytesResult = self._getState(StateType.TIME)
        return Conversion.uint8_array_to_uint32(bytesResult)
    
    
    
    
    """
    ---------------  UTIL  ---------------
    """
    
    
    def _writeToState(self):
        pass
    
    def _getState(self, stateType):
        """
        :param stateType: StateType
        """
        result = self.core.ble.setupSingleNotification(CSServices.CrownstoneService, CrownstoneCharacteristics.Result, lambda: self._requestState(stateType))

        resultPacket = ResultPacket(result)
        if resultPacket.valid:
            # the payload of the resultPacket is padded with stateType and ID at the beginning

            state = []
            for i in range(6, len(resultPacket.payload)):
                state.append(resultPacket.payload[i])

            return state
        else:
            raise CrownstoneException(CrownstoneError.INCORRECT_RESPONSE_LENGTH, "Result is invalid")

    def _requestState(self, stateType):
        self.core.ble.writeToCharacteristic(
            CSServices.CrownstoneService,
            CrownstoneCharacteristics.Control,
            ControlStateGetPacket(stateType).getPacket()
        )



class SystemBleTopics:
    rawAdvertisement = "rawAdvertisement"  # used to propagate verified state messages through the system
    abortScanning    = "abortScanning"  # used to propagate verified state messages through the system



"""
Simple BLE forever-scan example, that prints all the detected
LE advertisement packets, and prints a colored diff of data on data changes.
"""
import sys
import bluetooth._bluetooth as bluez

from bluetooth_utils import (toggle_device,
                             enable_le_scan, parse_le_advertising_events,
                             disable_le_scan, raw_packet_to_str)

dev_id = 0  # the bluetooth device is hci0
toggle_device(dev_id, True)

try:
    sock = bluez.hci_open_dev(dev_id)
except:
    print("Cannot open bluetooth device %i" % dev_id)
    raise

enable_le_scan(sock, filter_duplicates=False)

try:
    prev_data = None

    def le_advertise_packet_handler(mac, adv_type, data, rssi):
        global prev_data
        data_str = raw_packet_to_str(data)
        data_wo_rssi = (mac, data_str)
        print("BLE packet: %s %02x %s %d" % (mac, adv_type, data_str, rssi))
        if prev_data is not None:
            if data_wo_rssi != prev_data:
                # color differences with previous packet data
                sys.stdout.write(' ' * 35 + 'data_diff=')
                for c1, c2 in zip(data_str, prev_data[1]):
                    if c1 != c2:
                        sys.stdout.write('\033[0;33m' + c1 + '\033[m')
                    else:
                        sys.stdout.write(c1)
                sys.stdout.write('\n')

        prev_data = data_wo_rssi

    # Blocking call (the given handler will be called each time a new LE
    # advertisement packet is detected)
    parse_le_advertising_events(sock,
                                handler=le_advertise_packet_handler,
                                debug=False)
except KeyboardInterrupt:
    disable_le_scan(sock)



#!/usr/bin/env python3
from aioblescan import aioblescan

class AioCrownstoneParser(object):

    def __init__(self, hciEvent):
        self.hciEvent = hciEvent
        self.address = None
        self.rssi = None
        self.serviceUUID = None
        self.name = None
        self.valueText = None
        self.valueArr = None

        self.hasAdvertisement = False
        self.dataParser(self.hciEvent)

    def dataParser(self, data):
        for d in data.payload:
            if isinstance(d, aioblescan.HCI_CC_Event):
                pass
            elif isinstance(d, aioblescan.Byte):
                pass
                # print("Byte", d.name, d.val)
            elif isinstance(d, aioblescan.UIntByte):
                pass
                # print("UIntByte", d.name, d.val)
            elif isinstance(d, aioblescan.EnumByte):
                pass
                # print("EnumByte", d.name, d.val)
            elif isinstance(d, aioblescan.MACAddr):
                self.address = d.val
                # print("MACAddr", d.name, d.val)
            elif isinstance(d, aioblescan.IntByte):
                if d.name == 'rssi':
                    self.rssi = d.val
                # print("IntByte", d.name, d.val)
            elif isinstance(d, aioblescan.Itself):
                # print("Itself", d.name, d.val)
                self.valueText = d.val
                self.valueArr  = self.dataToBytes(d.val)
            elif isinstance(d, aioblescan.NBytes):
                self.serviceUUID = self.dataToNum(d.val)
            elif isinstance(d, aioblescan.String):
                self.name = d.val
                # print("String", d.name, d.val)
            elif isinstance(d, aioblescan.BitFieldByte):
                pass
                # print("BitFieldByte", d.name, d._val)
            elif isinstance(d, aioblescan.Adv_Data):
                self.dataParser(d)
            elif isinstance(d, aioblescan.HCI_LE_Meta_Event):
                self.dataParser(d)
            elif isinstance(d, aioblescan.HCI_LEM_Adv_Report):
                self.hasAdvertisement = True
                self.dataParser(d)
            else:
                pass
                # print(d)

    def dataToBytes(self, data):
        arr = []
        for b in data:
            arr.append(b)

        return arr

    def dataToNum(self, data):
        res = 0
        counter = len(data)-1
        for b in data:
            res += b << 8*counter
            counter -= 1
        return res


    def show(self):
        if self.hasAdvertisement:
            print("Address:", self.address, "RSSI:", self.rssi, "Length:", len(self.valueArr), "ServiceUUID:", hex(self.serviceUUID), "Data:", self.valueArr, "Name:", self.name)



from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics


class NormalModeChecker:

    def __init__(self, address, waitUntilInRequiredMode=False):
        self.address = address
        self.result = False
        self.waitUntilInRequiredMode = waitUntilInRequiredMode

    def handleAdvertisement(self, advertisement):
        if "serviceData" not in advertisement:
            return

        if advertisement["address"] != self.address:
            return

        self.result = not advertisement["serviceData"]["setupMode"]

        if not self.result and self.waitUntilInRequiredMode:
            pass
        else:
            BleEventBus.emit(SystemBleTopics.abortScanning, True)

    def getResult(self):
        return self.result




from crownstone_core.packets.Advertisement import Advertisement

from crownstone_ble.AioCrownstoneParser import AioCrownstoneParser
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics


class AioScanDelegate:
    
    def __init__(self, settings):
        self.settings = settings

    def handleDiscovery(self, hciEvent):
        parsedCrownstoneData = AioCrownstoneParser(hciEvent)
        self.parsePayload(parsedCrownstoneData)

    def parsePayload(self, parsedCrownstoneData):
        advertisement = Advertisement(
            parsedCrownstoneData.address,
            parsedCrownstoneData.rssi,
            parsedCrownstoneData.name,
            parsedCrownstoneData.valueArr,
            parsedCrownstoneData.serviceUUID
        )

        if advertisement.serviceData.opCode <= 5:
            advertisement.decrypt(self.settings.basicKey)
        elif advertisement.serviceData.opCode >= 7:
            advertisement.decrypt(self.settings.serviceDataKey)

        print("parsing a packet", advertisement.isCrownstoneFamily())
        if advertisement.isCrownstoneFamily():
            BleEventBus.emit(SystemBleTopics.rawAdvertisement, advertisement)


import time
import threading

AMOUNT_OF_REQUIRED_MATCHES = 2

class StoneAdvertisementTracker:
    
    def __init__(self, cleanupCallback):
        self.rssiHistory = None
        self.rssi = None
        self.name = ""
        self.address = None
        self.crownstoneId = 0
        self.lastUpdate = 0
        self.avgRssi = 0
        self.cleanupCallback = None
        self.uniqueIdentifier = 0
        self.verified = False
        self.dfu = False
        
        # config
        self.timeoutDuration = 4  # seconds
        self.rssiTimeoutDuration = 3  # seconds
        self.consecutiveMatches = 0
        
        self.timeoutTime = 0
        self.rssiTimeoutList = None
        self._lock = None
        
        self.rssiHistory = {}
        self.rssiTimeoutList = []
        self.cleanupCallback = cleanupCallback
        self._lock = threading.Lock()


    def tick(self):
        now = time.time()
        # check time in self.timeoutTime with current time
        if self.timeoutTime <= now:
            self.cleanupCallback()
            return

        # loop over rssi list
        remainingList = []
        for measurement in self.rssiTimeoutList:
            if measurement["timeoutTime"] <= now:
                del self.rssiHistory[measurement["key"]]
            else:
                remainingList.append(measurement)

        self.rssiTimeoutList = None
        self.rssiTimeoutList = remainingList
        self.avgRssi = self.calculateRssiAverage()



    def update(self, advertisement):
        self.name = advertisement.name
        self.address = advertisement.address
        self.avgRssi = advertisement.rssi
        if advertisement.isCrownstoneFamily():
            self.handlePayload(advertisement)
                
                
    def handlePayload(self, advertisement):
        # this field can be manipulated during the loop in calculate. To avoid this, we lock the threads for the duration of the loop
        with self._lock:
            self.rssi = advertisement.rssi
        
            now = time.time()
            
            self.lastUpdate = now
    
            self.rssiHistory[now] = self.rssi
        
            if advertisement.isInDFUMode():
                self.verified = True
                self.dfu = True
                self.consecutiveMatches = 0
            else:
                self.verify(advertisement.serviceData)
            
            self.avgRssi = self.calculateRssiAverage()
            self.timeoutTime = now + self.timeoutDuration
            self.rssiTimeoutList.append({"key": now, "timeoutTime": now + self.rssiTimeoutDuration})


    # check if we consistently get the ID of this crownstone.
    def verify(self, serviceData):
        if serviceData.isInSetupMode():
            self.verified = True
            self.consecutiveMatches = 0
        else:
            if not serviceData.dataReadyForUse:
                self.invalidateDevice(serviceData)
            else:
                if self.uniqueIdentifier != serviceData.uniqueIdentifier:
                    if serviceData.validation != 0 and serviceData.opCode == 5:
                        if serviceData.validation == 0xFA and serviceData.dataType != 1: # datatype 1 is the error packet
                            self.addValidMeasurement(serviceData)
                        elif serviceData.validation != 0xFA and serviceData.dataType != 1: # datatype 1 is the error packet
                            self.invalidateDevice(serviceData)
                    elif serviceData.validation != 0 and serviceData.opCode == 3:
                        if serviceData.dataType != 1:
                            if serviceData.validation == 0xFA:
                                self.addValidMeasurement(serviceData)
                            elif serviceData.validation != 0xFA:
                                self.invalidateDevice(serviceData)
                        else:
                            pass # do nothing, skip these messages (usually error states)
                    else:
                        if not serviceData.stateOfExternalCrownstone:
                            if serviceData.crownstoneId == self.crownstoneId:
                                self.addValidMeasurement(serviceData)
                            else:
                                self.invalidateDevice(serviceData)
        
        self.uniqueIdentifier = serviceData.uniqueIdentifier
        

    
    def addValidMeasurement(self, serviceData):
        if self.consecutiveMatches >= AMOUNT_OF_REQUIRED_MATCHES:
            self.verified = True
            self.consecutiveMatches = 0
        else:
            self.consecutiveMatches += 1
        
        self.crownstoneId = serviceData.crownstoneId



    def invalidateDevice(self, serviceData):
        if not serviceData.stateOfExternalCrownstone:
            self.crownstoneId = serviceData.crownstoneId
        
        self.consecutiveMatches = 0
        self.verified = False



    def calculateRssiAverage(self):
        count = 0
        total = 0

        keyList = list(self.rssiHistory)
        for key in keyList:
            total = total + self.rssiHistory[key]
            count += 1

        if count > 0:
            return total / float(count)
        else:
            return 0
        
        
        


class Gatherer:
    
    def __init__(self):
        self.deviceList = {}
        
    def handleAdvertisement(self, advertisement, validated=False):
        if "serviceData" not in advertisement:
            return
        
        rssi = float(advertisement["rssi"])
        if float(advertisement["rssi"]) >= 0:
            rssi = None
        
        if advertisement["address"] not in self.deviceList:
            self.deviceList[advertisement["address"]] = { "address": advertisement["address"].lower(), "setupMode": advertisement["serviceData"]["setupMode"], "validated": validated, "rssi": rssi }
        elif validated:
            self.deviceList[advertisement["address"]]["validated"] = True

        self.deviceList[advertisement["address"]]["setupMode"] = advertisement["serviceData"]["setupMode"]
        
        if self.deviceList[advertisement["address"]]["rssi"] is None:
            self.deviceList[advertisement["address"]]["rssi"]  = rssi
        else:
            self.deviceList[advertisement["address"]]["rssi"]  = 0.9*self.deviceList[advertisement["address"]]["rssi"] + 0.1*advertisement["rssi"]
        
    
    def getCollection(self):
        collectionArray = []
        for address, device in self.deviceList.items():
            collectionArray.append(device)
            
        return collectionArray


from enum import Enum


class BleError(Enum):
    CAN_NOT_FIND_SERVICE              = "CAN_NOT_FIND_SERVICE"
    CAN_NOT_FIND_CHACTERISTIC         = "CAN_NOT_FIND_CHACTERISTIC"
    CAN_NOT_GET_CHARACTERISTIC        = "CAN_NOT_GET_CHARACTERISTIC"
    CAN_NOT_FIND_CCCD                 = "CAN_NOT_FIND_CCCD"
    ABORT_NOTIFICATION_STREAM_W_ERROR = "ABORT_NOTIFICATION_STREAM_W_ERROR"
    NOTIFICATION_STREAM_TIMEOUT       = "NOTIFICATION_STREAM_TIMEOUT"
    NO_NOTIFICATION_DATA_RECEIVED     = "NO_NOTIFICATION_DATA_RECEIVED"
    INVALID_SESSION_NONCE             = "INVALID_SESSION_NONCE"
    INVALID_SESSION_DATA              = "INVALID_SESSION_DATA"
    INVALID_ENCRYPTION_PACKAGE        = "INVALID_ENCRYPTION_PACKAGE"
    INVALID_ENCRYPTION_USER_LEVEL     = "INVALID_ENCRYPTION_USER_LEVEL"
    COULD_NOT_VALIDATE_SESSION_NONCE  = "COULD_NOT_VALIDATE_SESSION_NONCE"
    COULD_NOT_READ_SESSION_NONCE      = "COULD_NOT_READ_SESSION_NONCE"
    NO_SESSION_NONCE_SET              = "NO_SESSION_NONCE_SET"
    NO_ENCRYPTION_KEYS_SET            = "NO_ENCRYPTION_KEYS_SET"
    ENCRYPTION_VALIDATION_FAILED      = "ENCRYPTION_VALIDATION_FAILED"
    SETUP_FAILED                      = "SETUP_FAILED"
    NOT_IN_RECOVERY_MODE              = "NOT_IN_RECOVERY_MODE"
    RECOVERY_MODE_DISABLED            = "RECOVERY_MODE_DISABLED"


from crownstone_core.Exceptions import CrownstoneError, CrownstoneBleException, CrownstoneException
from crownstone_core.core.modules.EncryptionSettings import EncryptionSettings
from crownstone_core.topics.Topics import Topics
from crownstone_core.util.JsonFileStore import JsonFileStore

from crownstone_ble.Exceptions import BleError
from crownstone_ble.constants import ScanBackends
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.core.ble_modules.BleHandler import BleHandler
from crownstone_ble.core.ble_modules.BroadcastHandler import BroadcastHandler
from crownstone_ble.core.ble_modules.ControlHandler import ControlHandler
from crownstone_ble.core.ble_modules.SetupHandler import SetupHandler
from crownstone_ble.core.ble_modules.StateHandler import StateHandler
from crownstone_ble.core.modules.Gatherer import Gatherer
from crownstone_ble.core.modules.NearestSelector import NearestSelector
from crownstone_ble.core.modules.NormalModeChecker import NormalModeChecker
from crownstone_ble.core.modules.RssiChecker import RssiChecker
from crownstone_ble.core.modules.SetupChecker import SetupChecker
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics


class CrownstoneBle:
    
    def __init__(self, hciIndex = 0, scanBackend = ScanBackends.Bluepy):
        self.settings  = EncryptionSettings()
        self.control   = ControlHandler(self)
        self.broadcast = BroadcastHandler(self, hciIndex)
        self.setup     = SetupHandler(self)
        self.state     = StateHandler(self)
        self.ble       = BleHandler(self.settings, hciIndex, scanBackend)
        
    def shutDown(self):
        self.ble.shutDown()
    
    def setSettings(self, adminKey, memberKey, basicKey, serviceDataKey, localizationKey, meshApplicationKey, meshNetworkKey, referenceId = "PythonLib"):
        self.settings.loadKeys(adminKey, memberKey, basicKey, serviceDataKey, localizationKey, meshApplicationKey, meshNetworkKey, referenceId)
        
        
    def loadSettingsFromFile(self, path):
        fileReader = JsonFileStore(path)
        data = fileReader.getData()
        
        if "admin" not in data:
            raise CrownstoneBleException(CrownstoneError.ADMIN_KEY_REQURED)
        if "member" not in data:
            raise CrownstoneBleException(CrownstoneError.MEMBER_KEY_REQUIRED)
        if "basic" not in data:
            raise CrownstoneBleException(CrownstoneError.BASIC_KEY_REQURED)
        if "serviceDataKey" not in data:
            raise CrownstoneBleException(CrownstoneError.SERVICE_DATA_KEY_REQUIRED)
        if "localizationKey" not in data:
            raise CrownstoneBleException(CrownstoneError.LOCALIZATION_KEY_REQUIRED)
        if "meshApplicationKey" not in data:
            raise CrownstoneBleException(CrownstoneError.MESH_APP_KEY)
        if "meshNetworkKey" not in data:
            raise CrownstoneBleException(CrownstoneError.MESH_NETWORK_KEY)

        self.setSettings(data["admin"], data["member"], data["basic"], data["serviceDataKey"], data["localizationKey"], data["meshApplicationKey"], data["meshNetworkKey"])
        

    def connect(self, address, ignoreEncryption=False, connectionSettings=None):
        self.ble.connect(address, connectionSettings)
        if not ignoreEncryption:
            try:
                self.control.getAndSetSessionNone()
            except CrownstoneBleException as err:
                # the only relevant error here is this one. If it is any other, the Crownstone is in the wrong mode
                if err.type is BleError.COULD_NOT_VALIDATE_SESSION_NONCE:
                    raise err


    def setupCrownstone(self, address, sphereId, crownstoneId, meshDeviceKey, ibeaconUUID, ibeaconMajor, ibeaconMinor):
        self.connect(address)
        self.setup.setup(address, sphereId, crownstoneId, meshDeviceKey, ibeaconUUID, ibeaconMajor, ibeaconMinor)
        self.disconnect()
        
    
    def disconnect(self):
        self.ble.disconnect()
    
    def startScanning(self, scanDuration=3):
        self.ble.startScanning(scanDuration)

    def stopScanning(self):
        self.ble.stopScanning()

    def getCrownstonesByScanning(self, scanDuration=3):
        gatherer = Gatherer()
    
        subscriptionIdValidated = BleEventBus.subscribe(Topics.advertisement,             lambda advertisementData: gatherer.handleAdvertisement(advertisementData, True)          )
        subscriptionIdAll       = BleEventBus.subscribe(SystemBleTopics.rawAdvertisement, lambda advertisement: gatherer.handleAdvertisement(advertisement.getDictionary(), False) )
    
        self.ble.startScanning(scanDuration=scanDuration)
    
        BleEventBus.unsubscribe(subscriptionIdValidated)
        BleEventBus.unsubscribe(subscriptionIdAll)
        
        return gatherer.getCollection()

    def isCrownstoneInSetupMode(self, address, scanDuration=3, waitUntilInRequiredMode=False):
        # print("Checking if it is in setup mode, address", address)
        checker = SetupChecker(address, waitUntilInRequiredMode)
        subscriptionId = BleEventBus.subscribe(Topics.advertisement, checker.handleAdvertisement)

        self.ble.startScanning(scanDuration=scanDuration)

        BleEventBus.unsubscribe(subscriptionId)

        return checker.getResult()

    def isCrownstoneInNormalMode(self, address, scanDuration=3, waitUntilInRequiredMode=False):
        # print("Checking if it is in normal mode, address", address)
        checker = NormalModeChecker(address, waitUntilInRequiredMode)
        subscriptionId = BleEventBus.subscribe(SystemBleTopics.rawAdvertisement, lambda advertisement: checker.handleAdvertisement(advertisement.getDictionary()))

        self.ble.startScanning(scanDuration=scanDuration)

        BleEventBus.unsubscribe(subscriptionId)

        return checker.getResult()

    def getRssiAverage(self, address, scanDuration=3):
        checker = RssiChecker(address)
        subscriptionId = BleEventBus.subscribe(SystemBleTopics.rawAdvertisement, lambda advertisement: checker.handleAdvertisement(advertisement.getDictionary()))

        self.ble.startScanning(scanDuration=scanDuration)

        BleEventBus.unsubscribe(subscriptionId)

        return checker.getResult()

    def getNearestCrownstone(self, rssiAtLeast=-100, scanDuration=3, returnFirstAcceptable=False, addressesToExclude=[]):
        return self._getNearest(False, rssiAtLeast, scanDuration, returnFirstAcceptable, False, addressesToExclude)
    
    
    def getNearestValidatedCrownstone(self, rssiAtLeast=-100, scanDuration=3, returnFirstAcceptable=False, addressesToExclude=[]):
        return self._getNearest(False, rssiAtLeast, scanDuration, returnFirstAcceptable, True, addressesToExclude)
    
    
    def getNearestSetupCrownstone(self, rssiAtLeast=-100, scanDuration=3, returnFirstAcceptable=False, addressesToExclude=[]):
        return self._getNearest(True, rssiAtLeast, scanDuration, returnFirstAcceptable, True, addressesToExclude)
    

    def _getNearest(self, setup, rssiAtLeast, scanDuration, returnFirstAcceptable, validated, addressesToExclude):
        addressesToExcludeSet = set()
        for data in addressesToExclude:
            if isinstance(data, dict):
                if "address" in data:
                    addressesToExcludeSet.add(data["address"].lower())
                else:
                    raise CrownstoneException(CrownstoneError.INVALID_ADDRESS, "Addresses to Exclude is either an array of addresses (like 'f7:19:a4:ef:ea:f6') or an array of dicts with the field 'address'")
            else:
                addressesToExcludeSet.add(data.lower())

        selector = NearestSelector(setup, rssiAtLeast, returnFirstAcceptable, addressesToExcludeSet)
    
        topic = Topics.advertisement
        if not validated:
            topic = SystemBleTopics.rawAdvertisement
            subscriptionId = BleEventBus.subscribe(topic, lambda advertisement: selector.handleAdvertisement(advertisement.getDictionary()))
        else:
            subscriptionId = BleEventBus.subscribe(topic, lambda advertisementData: selector.handleAdvertisement(advertisementData))
    

        
        self.ble.startScanning(scanDuration=scanDuration)
    
        BleEventBus.unsubscribe(subscriptionId)
        
        return selector.getNearest()


"""
Simple BLE advertisement example
"""
from time import sleep
import bluetooth._bluetooth as bluez

from bluetooth_utils import (toggle_device, start_le_advertising,
                             stop_le_advertising)

dev_id = 0  # the bluetooth device is hci0
toggle_device(dev_id, True)

try:
    sock = bluez.hci_open_dev(dev_id)
except:
    print("Cannot open bluetooth device %i" % dev_id)
    raise

try:
    start_le_advertising(sock,
                         min_interval=2000, max_interval=2000,
                         data=(0x11, 0x22, 0x33) + (0,) * 28)
    while True:
        sleep(2)
except:
    stop_le_advertising(sock)
    raise




class RssiChecker:

    def __init__(self, address):
        self.address = address
        self.result = []

    def handleAdvertisement(self, advertisement):
        if "serviceData" not in advertisement:
            return

        if advertisement["address"] != self.address:
            return

        self.result.append(advertisement["rssi"])


    def getResult(self):
        if len(self.result) == 0:
            return None

        sum = 0
        for res in self.result:
            sum += res
        return sum/len(self.result)





#!/usr/bin/env python3

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='crownstone-ble',
    version='0.5.3',
    packages=find_packages(exclude=["examples","testing"]),
    author="Crownstone B.V.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/crownstone/crownstone-lib-python-ble",
    install_requires=list(package.strip() for package in open('requirements.txt')),
    classifiers=[
        'Programming Language :: Python :: 3.7'
    ],
    python_requires='>=3.7',
)


import bluetooth._bluetooth as bluez
from time import sleep

from crownstone_core import Conversion
from crownstone_core.core.modules.EncryptionSettings import UserLevel
from crownstone_core.protocol.BluenetTypes import BroadcastTypes
from crownstone_core.util.EncryptionHandler import EncryptionHandler
from crownstone_core.util.RC5 import RC5

from crownstone_ble.thirdParty.py_bluetooth_utils.bluetooth_utils import start_le_advertising, stop_le_advertising


class BroadcastHandler:

    def __init__(self, bluetoothCore, hciIndex):
        self.core = bluetoothCore
        self.broadcaster = Broadcaster(hciIndex, self.core.settings)

    def switchCrownstone(self, sphereUID, crownstoneId, switchState):
        """
        switchState is in the range of [0 .. 1]
        """
        convertedSwitchState = int(min(1, max(0, switchState)) * 100)
        self._switchCrownstone(sphereUID, crownstoneId, convertedSwitchState)

    def turnOnCrownstone(self, sphereUID, crownstoneId):
        self._switchCrownstone(sphereUID, crownstoneId, 255)

    def turnOffCrownstone(self, sphereUID, crownstoneId):
        self._switchCrownstone(sphereUID, crownstoneId, 0)

    def _switchCrownstone(self, sphereUID, crownstoneId, switchState):
        """
        Switchstate here is conforming to the Crownstone protocol, range 0..255
        """
        stoneSwitchPackets = [[crownstoneId, switchState]]

        packet = [BroadcastTypes.MULTI_SWITCH.value, len(stoneSwitchPackets) * 2]
        for switchPacket in stoneSwitchPackets:
            packet += switchPacket

        self.broadcaster.cast(sphereUID, packet)



class Broadcaster:

    def __init__(self, hciIndex, settings):
        self.broadcastCounter = 0
        self.hciIndex = hciIndex
        self.deviceToken = 0xf0
        self.settings = settings

    def cast(self, sphereUID, payload):
        uint16ServiceNumbers = BroadcastProtocol.getUInt16ServiceNumbers(
            self.broadcastCounter,
            self.deviceToken,
            sphereUID,
            UserLevel.basic,
            self.settings.localizationKey
        )

        commandPayload = BroadcastProtocol.getEncryptedServiceUUID(
            self.settings.basicKey,
            payload,
            Conversion.uint16_array_to_uint8_array(uint16ServiceNumbers),
            Conversion.uint32_to_uint8_array(0xCAFEBABE) # this is a shortcut, we will use timestamps at firmware v5 so it is immune to repeat attacks.
        )


        try:
            socket = bluez.hci_open_dev(self.hciIndex)
        except:
            print("Cannot open bluetooth device %i" % self.hciIndex)
            raise
        payload = [
            0x02, 0x01, 0x1A, # prefix
            0x11, 0x07, *list(commandPayload),
            0x09, 0x03, *Conversion.uint16_array_to_uint8_array(uint16ServiceNumbers),
        ]

        payload += [0] * (31-len(payload))

        try:
            start_le_advertising(socket, min_interval=20, max_interval=40, data=payload)
            sleep(1.5)
        except:
            stop_le_advertising(socket)
            raise



class BroadcastProtocol:

    """/**
     * Payload is 12 bytes, this method will add the validation and encrypt the thing
     **/
     """
    @staticmethod
    def getEncryptedServiceUUID(key, data, IV, validationNonce):
        encryptedData = EncryptionHandler.encryptBroadcast(data, key, IV, validationNonce)
        return encryptedData


    @staticmethod
    def getUInt16ServiceNumbers(broadcastCounter, deviceToken, sphereUID, accessLevel, key):
        result = []

        firstPart = broadcastCounter << 8

        protocolVersion = 0

        RC5Component = BroadcastProtocol.getRC5Payload(firstPart, key)
        result.append(BroadcastProtocol._constructProtocolBlock(protocolVersion, sphereUID, accessLevel))
        result.append(BroadcastProtocol._getFirstRC5Block(deviceToken, RC5Component))
        result.append(BroadcastProtocol._getSecondRC5Block(RC5Component))
        result.append(BroadcastProtocol._getThirdRC5Block(RC5Component))

        return result


    """ 
     * This is an UInt32 which will be encrypted
     * | ----------- First Part ---------- |
     * | Counter         | Reserved        | LocationId  | Profile Index | RSSI Calibration | Flag: t2t enabled | Flag: ignore for behaviour | Flag: reserved |
     * | 0 0 0 0 0 0 0 0 | 0 0 0 0 0 0 0 0 | 0 0 0 0 0 0 | 0 0 0         | 0 0 0 0          | 0                 | 0                          | 0              |
     * | 8b              | 8b              | 6b          | 3b            | 4b               | 1b                | 1b                         | 1b             |
     *
     * The first part will be a validation nonce in the background.
     *
     """
    @staticmethod
    def getRC5Payload(firstPart, key):
        rc5Payload = 0

        rc5Payload += firstPart << 16

        rc5Payload += (0 & 0x0000003F) << 10 # locationId, ignore

        rc5Payload += (0 & 0x00000007) << 7 # profile Index, ignore

        rc5Payload += (8 & 0x0000000F) << 3 # rssi offset, ignore (value 0 results in 8)

        # tap to toggle
        rc5Payload += 0 << 2 #, ignore
        
        # ignore for Behaviour
        rc5Payload += 1 << 1

        rc5Encryptor = RC5(key,32,12)
        return rc5Encryptor.encrypt(rc5Payload)


    """/**
    * This is an UInt16 is constructed from an index flag, then a protocol,  the Sphere passkey and the access level
    *
    * | Index |  Protocol version |  Sphere UID      |  Access Level |
    * | 0 0   |  0 0 0            |  0 0 0 0 0 0 0 0 |  0 0 0        |
    * | 2b    |  3b               |  8b              |  3b           |
    *
    """
    @staticmethod
    def _constructProtocolBlock(protocolVersion, spherePasskey, accessLevel):
        block = 0

        block += (protocolVersion & 0x0007) << 11
        if spherePasskey is not None:
            block += (spherePasskey & 0x00FF) << 3
        
        block += accessLevel.value & 0x0007

        return block


    """ 
     * TThe first chunk of RC5 data and reserved chunk of public bits
     *
     * | Index |  Reserved  |  Device Token     |  First chunk of RC5Data  |
     * | 0 1   |  0 0       |  0 0 0 0 0 0 0 0  |  0 0 0 0   |
     * | 2b    |  2b        |  8b                     |  4b          |
     *
    """
    @staticmethod
    def _getFirstRC5Block(deviceToken, RC5):
        block = 0

        block += 1 << 14 # place index

        # place device token
        block += deviceToken << 4

        block += (RC5 & 0xF0000000) >> 28

        return block
    

    """ 
     *
     * | Index |  RC chunk 2                       |
     * | 1 0   |  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  |
     * | 2b    |  14b                              |
     *
    """
    @staticmethod
    def _getSecondRC5Block(RC5):
        block = 0

        block += 1 << 15 # place index
        block += (RC5 & 0x0FFFC000) >> 14

        return block
    

    """ 
     *
     * | Index |  RC chunk 3                       |
     * | 1 1   |  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0  |
     * | 2b    |  14b                              |
     *
    """
    @staticmethod
    def _getThirdRC5Block(RC5):
        block = 0

        block += 3 << 14 # place index
        block += RC5 & 0x00003FFF

        return block



from crownstone_ble.constants import ScanBackends
from crownstone_ble.core.CrownstoneBle import CrownstoneBle
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.BleTopics import BleTopics
